#------------------------------------------------------------------------------------------------
# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#------------------------------------------------------------------------------------------------
# THIS PLUGIN WAS DEVELOPED AND TESTED UNDER WINDOWS VISTA ONLY AND MAY OR MAY NOT WORK ON A MAC.
# THIS IS ANOTHER PRODUCT OF MY IMAGINATION AND NOT BASED ON BUILDING OR OTHER ACCEPTED PRACTICES.
#------------------------------------------------------------------------------------------------
#    Name:	Generate Ceiling Grid V
#      By:	sdmitch
#   Usage:	Generates a drop ceiling from a face.  Select face then run plugin. Choose tile size,
#						2x2',2x4',4x4'; origin, corner or center; rotate 90, yes or no; add tiles, yes or no.
#		 Note:	I have tried to make the plugin as flexible as possible regarding the shape of the face
#						but it is next to impossible to code for every conceivable circumstance.
#						Interior columns or walls that create "holes" in the shape are acceptable.
#						If Make Tiles is Yes, grouped individual tiles will be created and painted with the
#						current material. If No, the selected face will be painted with the current material.
#    Date:	Mar 2012
#------------------------------------------------------------------------------------------------

require 'Sketchup'

if !file_loaded?(File.basename(__FILE__))
	UI.menu("Plugins").add_item("Ceiling Grid") { SDM_GCGV.gen_grid }
	file_loaded(File.basename(__FILE__))
end

module SDM_GCGV
	
	def self.gen_grid
		mod = Sketchup.active_model
		ent = mod.active_entities
		sel = mod.selection
		unless sel.empty? || sel.first.class!=Sketchup::Face
			@tile_size = "2X4" if !@tile_size; @start_pt = "Corner" if !@start_pt; @rotate = "No" if !@rotate; @add_tiles = "No" if !@add_tiles
			prompts = ["   Tile Size:","   Origin Pt:","  Rotate 90:","Make Tiles:"]
			options = ["2X2|2X4|4X4","Corner|Center","Yes|No","Yes|No"]
			default = [@tile_size,@start_pt,@rotate,@add_tiles]
			ans=UI.inputbox(prompts,default,options, "Generate Suspended Ceiling Grid")
			if !ans then return end
			@tile_size=ans[0]; @start_pt = ans[1]; @rotate = ans[2]; @add_tiles = ans[3]
			mod.start_operation "Ceiling Grid"
			egrp = ent.add_group;@ge = egrp.entities;
			face = sel.first; face.reverse! if face.normal.z > 0.0
			pts=face.outer_loop.vertices.collect{|v| v.position}
			case @tile_size
				when "2X2" ; dx = 24.125.inch; dy = 24.125.inch;
				when "2X4" ; dx = 48.125.inch; dy = 24.125.inch;
				when "4X4" ; dx = 48.125.inch; dy = 48.125.inch;
				else
					UI.messagebox "tile_size #{@tile_size} not found"
					return
			end
			if self.face_to_grid(face,pts,dx,dy)
				### convert boundary edges in to "L" members
				for loop in face.loops
					edges=loop.edges
					edges.each{|e| self.create_cgm(face,e,"E")}
				end
				if @add_tiles == "Yes"
					ent.erase_entities(face)
					existing_faces=[];created_faces=[];tile_faces=[];sel.clear
					ent.each{|f| existing_faces<<f if f.is_a?(Sketchup::Face)}
					dump=egrp.explode;dump.each{|c| sel.add c if c.is_a?(Sketchup::ComponentInstance)}
					cgrp = ent.add_group sel; cgrp.name = "T-Grid";
					dump.each{|e| e.find_faces if e.is_a?(Sketchup::Edge)}
					ent.each{|f| created_faces<<f if f.is_a?(Sketchup::Face) && !existing_faces.include?(f)}
					created_faces.each{|f| existing_faces<<f;f.reverse! if f.normal.z > 0.0;self.ct_offset(f,-0.125)}
					ent.each{|f| tile_faces<<f if f.is_a?(Sketchup::Face) && !existing_faces.include?(f)}
					mat = mod.materials.current;sel.clear;tile_faces.each{|f| sel.add f}
					tgrp = ent.add_group sel; tgrp.name = "Tiles"; tge = tgrp.entities
					tge.each{|f| (f.pushpull(-0.5);f.material=mat) if f.is_a?(Sketchup::Face) && tile_faces.include?(f)}
				else
					tgrp=ent.add_group face;face.pushpull(-0.5);face.material=mod.materials.current;sel.clear
					tgrp.name = "Ceiling"; egrp.name = "T-Grid"
				end
				mod.commit_operation
			end
		else
			UI.messagebox "Select a face"
		end
	end
	
	def self.face_to_grid(face,pts,dx,dy)
		mod = Sketchup.active_model
		ent = mod.active_entities
		ndx=1; # assume regular 4 sided rectangle
		if pts.length > 4
			longest_side = 0.0; #find the longest pair of adjacent sides
			for i in -1...pts.length-1
				dist=pts[i-1].distance(pts[i]) + pts[i].distance(pts[i+1])
				if dist > longest_side
					longest_side = dist; # save longer distance
					ndx = i; # save index of common point
				end
			end
		end
		d1=pts[ndx].distance(pts[ndx+1]); # length of X side
		d2=pts[ndx].distance(pts[ndx-1]); # length of Y side
		dmax = [dx,dy].max; ndx -= 1 if d2>d1
		if (face.bounds.diagonal >= dmax); # make sure the rectangle is big enough to sub-divide
			nx=(face.bounds.diagonal*2/dx).ceil;ny=(face.bounds.diagonal*2/dy).ceil
			if @rotate == "No"
				v1=pts[ndx].vector_to(pts[ndx+1]); # vector of X side
				v2=pts[ndx].vector_to(pts[ndx-1]); v3 = Geom::Vector3d.new(v1.y,-v1.x,0)
				v2.angle_between(v3) > Math::PI/2 ? v2 = v3.reverse : v2 = v3
			else
				v1=pts[ndx].vector_to(pts[ndx-1]); # vector of X side
				v2=pts[ndx].vector_to(pts[ndx+1]); v3 = Geom::Vector3d.new(v2.y,-v2.x,0)
				v1.angle_between(v3) > Math::PI/2 ? v1 = v3.reverse : v1 = v3
			end
			edges=[]; edges=face.edges.to_a;
			data=[]; row=[]; @yaxis = v2; # vector of Y side
			if @start_pt == "Corner"
				tx = [nx/4+1,3].max; ty =[ny/4+1,3].max
				pt0 = pts[ndx].offset(v1,-dx*tx).offset(v2,-dy*ty);#puts "corner pt0=#{pt0}"
			else
				tx = [nx/2,3].max; ty =[ny/2,3].max
				pt0 = face.bounds.center.offset(v1,-dx*tx).offset(v2,-dy*ty);#puts "center pt0=#{pt0}"
			end
			# Compute all end points
			for i in 0..ny
				row[0] = pt0
				for j in 1..nx
					row[j]=row[j-1].offset(v1,dx)
				end
				data[i]=row; row=[]
				pt0.offset!(v2,dy)
			end
####### Draw the "horizontal" lines ##################################################
			for i in 1..ny
				for j in 1..nx
					start_pt_outside = (face.classify_point(data[i][j-1]) == Sketchup::Face::PointOutside)
					end_pt_outside = (face.classify_point(data[i][j]) == Sketchup::Face::PointOutside)
					ent.add_cpoint(data[i][j]) if end_pt_outside && $sdm_debug
					if start_pt_outside && !end_pt_outside
						if [2,4].include?(face.classify_point(data[i][j]))
							pt0 = data[i][j]
						else
							pt0 = self.find_intersection(data[i][j-1],data[i][j],edges)
						end
					elsif end_pt_outside && !start_pt_outside
						pt1 = self.find_intersection(data[i][j-1],data[i][j],edges)
						@rotate == "Yes" ? pt2 = pts[ndx-1] : pt2 = pts[ndx+1]
						if pt1 && (!self.is_between(pt0,pts[ndx],pt2) || !self.is_between(pt1,pts[ndx],pt2))
							edge=@ge.add_line(pt0,pt1)
							self.create_cgm(face,edge,"T")
						end
					elsif !start_pt_outside && !end_pt_outside # start and end both on face
						if ![2,4].include?(face.classify_point(data[i][j-1]))  || ![2,4].include?(face.classify_point(data[i][j]))
							ray = [data[i][j-1],v1]; hit = mod.raytest(ray)
							if hit && [1,2,4].include?(face.classify_point(hit[0]))
								if self.is_between(hit[0],data[i][j-1],data[i][j])
									edge = @ge.add_line(pt0,hit[0])
									self.create_cgm(face,edge,"T")
									ray = [hit[0],v1]; hit = mod.raytest(ray)
									if hit && [1,2,4].include?(face.classify_point(hit[0]))
										pt0 = hit[0] if self.is_between(hit[0],data[i][j-1],data[i][j])
									end
								end
							end
						else
							pt0 = data[i][j-1]; 
						end
					elsif start_pt_outside && end_pt_outside # start and end both off the face
						ray = [data[i][j-1],v1]; hit = mod.raytest(ray); p0 = nil; p1 = nil
						if hit && [1,2,4].include?(face.classify_point(hit[0]))
							if self.is_between(hit[0],data[i][j-1],data[i][j])
								p0 = hit[0]; ray=[p0,v1]; hit = mod.raytest(ray)
								if hit && [1,2,4].include?(face.classify_point(hit[0]))
									p1 = hit[0] if self.is_between(hit[0],data[i][j-1],data[i][j])
									if p0 and p1
										edge = @ge.add_line(p0,p1)
										self.create_cgm(face,edge,"T")
									end
								end
							end
						end
					end
				end
			end
####### Draw the "vertical" lines #####################################################
			for j in 1..nx
				for i in 1..ny
					start_pt_outside = (face.classify_point(data[i-1][j]) == Sketchup::Face::PointOutside)
					end_pt_outside = (face.classify_point(data[i][j]) == Sketchup::Face::PointOutside)
					if start_pt_outside && !end_pt_outside
						if ![2,4].include?(face.classify_point(data[i][j]))
							poi=self.find_intersection(data[i-1][j],data[i][j],edges)
							if poi
								edge=@ge.add_line(poi,data[i][j])
								self.create_cgm(face,edge,"T")
							end
						end
					elsif !start_pt_outside && end_pt_outside
						if ![2,4].include?(face.classify_point(data[i-1][j]))
							poi=self.find_intersection(data[i-1][j],data[i][j],edges)
							if poi
								edge=@ge.add_line(data[i-1][j],poi)
								self.create_cgm(face,edge,"T")
							end
						end
					elsif !start_pt_outside && !end_pt_outside # start and end both on face
						if ![2,4].include?(face.classify_point(data[i-1][j]))  || ![2,4].include?(face.classify_point(data[i][j]))
							ray = [data[i-1][j],v2]; hit = mod.raytest(ray)
							if hit && [1,2,4].include?(face.classify_point(hit[0]))
								if edges.include?(hit[1][0]) && self.is_between(hit[0],data[i-1][j],data[i][j])
									edge = @ge.add_line(data[i-1][j],hit[0])
									self.create_cgm(face,edge,"T")
									ray=[hit[0],v2]; hit = mod.raytest(ray)
									if hit && [1,2,4].include?(face.classify_point(hit[0]))
										if edges.include?(hit[1][0]) && self.is_between(hit[0],data[i-1][j],data[i][j])
											edge = @ge.add_line(hit[0],data[i][j])
											self.create_cgm(face,edge,"T")
										end
									end
								else
									edge = @ge.add_line(data[i-1][j],data[i][j])
									self.create_cgm(face,edge,"T")
								end
							end
						end
					elsif start_pt_outside && end_pt_outside # start and end both off the face
						ray = [data[i-1][j],v2]; hit = mod.raytest(ray); p0 = nil
						if hit && [1,2,4].include?(face.classify_point(hit[0]))
							if edges.include?(hit[1][0]) && self.is_between(hit[0],data[i-1][j],data[i][j])
								p0 = hit[0]; ray=[p0,v2]; hit = mod.raytest(ray)
								if hit && [1,2,4].include?(face.classify_point(hit[0]))
									if edges.include?(hit[1][0]) && self.is_between(hit[0],data[i-1][j],data[i][j])
										if p0
											edge = @ge.add_line(p0,hit[0])
											self.create_cgm(face,edge,"T")
										end
									end
								end
							end
						end
					end
				end
			end
			return true
		else
			UI.messagebox "This face is to small for #{@tile_size} tiles";return false
		end
	end
	
	def self.create_cgm(face,edge,type)
		mod = Sketchup.active_model
		ent = mod.active_entities
		sel = mod.selection
		pt0 = edge.start.position; pt1 = edge.end.position
		if type == "E" && edge.reversed_in?(face)
			pt0 = edge.end.position; pt1 = edge.start.position
		end
		vec = pt0.vector_to(pt1).normalize; ppd = pt0.distance(pt1)
		unless vec.length==0
			cgm_id = "#{type}X #{ppd}"
			cgm_id = "#{type}Y #{ppd}" if vec.parallel?(@yaxis)
			defs = mod.definitions;cgm = nil
			defs.each{|d| cgm=d if d.name==cgm_id}
			if !cgm
				grp = ent.add_group; ge = grp.entities
				xa,ya,za = Z_AXIS.axes;origin = Geom::Point3d.new
				if type == "T"
					p0 = origin.offset(za,0.0625.inch)
					p1 = p0.offset(xa,-0.0625.inch)
					p2 = p1.offset(ya,+1.6250.inch)
					p3 = p2.offset(xa,+0.1250.inch)
					p4 = p3.offset(ya,-1.6250.inch)
					f = ge.add_face(p1,p2,p3,p4)
					f.pushpull -(ppd-0.125.inch)
					p0 = origin.offset(za,0.5.inch)
					p1 = p0.offset(xa,+0.5000.inch)
					p2 = p1.offset(ya,-0.1250.inch)
					p3 = p2.offset(xa,-1.0000.inch)
					p4 = p3.offset(ya,+0.1250.inch)
					f = ge.add_face(p1,p2,p3,p4)
					f.pushpull -(ppd-1.0.inch)
				elsif type == "E"
					p1 = origin.offset(ya,-0.1250)
					p2 = p1.offset(ya,+1.7500)
					p3 = p2.offset(xa,-0.0625)
					p4 = p3.offset(ya,-1.6250)
					p5 = p4.offset(xa,-0.4375)
					p6 = p5.offset(ya,-0.1250)
					f = ge.add_face(p1,p2,p3,p4,p5,p6)
					f.pushpull -ppd;
				end
				comp = grp.to_component
				comp.definition.name = cgm_id
				cgm = comp.definition
				comp.erase!
			end
			tr = Geom::Transformation.new(pt0,vec)
			@ge.add_instance(cgm,tr)
		end
	end
	
	def self.find_intersection(p0,p1,edges)
		ln0=[p0,p0.vector_to(p1).normalize]
		for e in edges
			ln1=e.line; pp=Geom.intersect_line_line(ln0,ln1)
			if pp && pp != p0
				if self.is_between(pp,e.start.position,e.end.position)
					if self.is_between(pp,p0,p1)
						return pp
					end
				end
			end
		end
		return nil
	end
	
	def  self.is_between(p0,p1,p2)
	# --------------------------------------------------------------------------------
	#	Determine if point p0 is on the line and between end points p1 and p2
	# --------------------------------------------------------------------------------
		line=[p1,p1.vector_to(p2).normalize]
		return false if !p0.on_line? line
		d1=p0.distance p1;d2=p0.distance p2;d3=p1.distance p2 
		return true if d1 <= d3 && d2 <= d3
		return false # p0 is not on line and between p1 and p2
	end

	def self.ct_offset(face,dist)
		mod = Sketchup.active_model
		ent = mod.active_entities
		sel = mod.selection
# Code copied from offset.rb
# Copyright 2004,2005,2006,2009 by Rick Wilson - All Rights Reserved
		begin
			pi = Math::PI
			if (not ((dist.class==Fixnum || dist.class==Float || dist.class==Length) && dist!=0))
				return nil
			end
			verts=face.outer_loop.vertices;pts = [];vecs = []
			# verts.reverse! if face.outer_loop.edges[0].reversed_in?(face)
			# CREATE ARRAY pts OF OFFSET POINTS FROM FACE
			0.upto(verts.length-1) do |a|
				vec1 = (verts[a].position-verts[a-(verts.length-1)].position).normalize
				vec2 = (verts[a].position-verts[a-1].position).normalize
				vec3 = (vec1+vec2).normalize
				if vec3.valid?
					ang = vec1.angle_between(vec2)/2
					ang = pi/2 if vec1.parallel?(vec2)
					vec3.length = dist/Math::sin(ang)
					vecs << vec3
					t = Geom::Transformation.new(vec3)
					if pts.length > 0
						vec4 = pts.last.vector_to(verts[a].position.transform(t))
						if vec4.valid?
							unless (vec2.parallel?(vec4))
								t = Geom::Transformation.new(vec3.reverse)
							end
						end
					end
					pts.push(verts[a].position.transform(t))
				else
					puts "vec3 is invalid"
				end
			end
			# CHECK FOR DUPLICATE POINTS IN pts ARRAY
			duplicates = []
			pts.each_index do |a|
				pts.each_index do |b|
					next if b==a
					duplicates<<b if pts[a]===pts[b]
				end
				break if a==pts.length-1
			end
			duplicates.reverse.each{|a| pts.delete(pts[a])}
			# CREATE FACE FROM POINTS IN pts ARRAY
			(pts.length > 2) ? (ent.add_face(pts)) : (return nil)
		rescue
			puts "#{face} did not offset: #{pts}"
			return nil
		end
	end
end
